import logging
from sqlite3 import *
from typing import (
    List,
    Union,
    Tuple,
    AnyStr, Type
)

from prettytable import PrettyTable

from .config import *
from .constants import *
from .exceptions import *
from .models import *
from .models import Table
from .testing import *

logging.basicConfig(format='%(asctime)s - INFO - %(message)s', level=logging.INFO)

__version__ = '0.1.4.13'
__author__ = 'Murodov Azizmurod'


class Database:
    """
    ORMX Database Object
    Attributes
    ----------
    __path : PathType
        Local __str__ to database (PathType)
    """

    def __init__(self, path):
        """
        Initialization
        Parameters
        ----------
        path : PathType
            Local __str__ to database (PathType)
        """
        self.conn = connect(path, check_same_thread=False, detect_types=PARSE_DECLTYPES)
        self.cur = self.conn.cursor()
        self.config = Config()

    def __new__(cls, *args, **kwargs):
        instance = super(Database, cls).__new__(cls)
        logging.info('Database connected!')
        return instance

    def __getitem__(self, key) -> ListQuery:
        return ListQuery(key, self)

    def _execute(self, sql: str, params: AnyStr = None) -> Cursor:
        """
        Execute any SQL-script whit (or without) values, or execute SQLRequest
        Parameters
        ----------
        sql : AnyStr
            single SQLite script, might contains placeholders
        params : AnyStr
            Values for placeholders if script contains it
        Returns
        ----------
        Union[List, None]
            Database answer if it has
        """
        try:
            if params:
                return self.conn.execute(sql, params)
            return self.conn.execute(sql)
        except OperationalError as e:
            pass

    @property
    def tables(self) -> List:
        return [x[0] for x in self._execute(SELECT_TABLES_SQL).fetchall()]

    def create(self, table: Table) -> None:
        """
        Creates table scheme
        :params
            table: Table Object
        :return:
            List or AnyStr
        """
        if len(table._get_create_sql()) == 2:
            try:
                self._execute(table._get_create_sql()[0])
                self._execute(table._get_create_sql()[1])
            except:
                pass
        else:
            self._execute(table._get_create_sql())

    def drop(self, table: Type[Table], if_exists=False) -> None:
        sql = table._get_drop_sql(exp=if_exists)
        self._execute(sql)
        self.conn.commit()

    def save(self, instance: Table):
        """
        Inserts data to `table`
        :params
            instance: Copy of Table Object
        :return:
            List of Table Objects
        """
        sql, values = instance._get_insert_sql()
        cursor = self._execute(sql, values)
        instance._data['id'] = cursor.lastrowid
        self.conn.commit()

    def all(self, table: Table,
            order_by: tuple = None,
            limit: list = None,
            where: list = None,
            fields: list = None,
            pretty_table: bool = False) -> List[Table]:
        """
        Returns all rows from `table`
        :params
            table: Table Object that will used
            order_by: name of column and sorting type. ex. ('title', ASC)
            limit: list of integers. ex. [10, 2] -> LIMIT 10 OFFSET 2
            where: list of filters.
                ex: ['column_name', 'condition', 'value']
                    or
                    [
                        ('name', '==', 'title'),
                        AND,
                        ('draft', '!=', 0),
                        OR
                        ('published', '>=', datetime.now())
                    ]
            fields: list of fields which will be returned. ex. ['title', 'published']
            pretty_table: bool. Printing query with PrettyTable
        :return:
            List of Table Objects
        """
        result = []
        sql, fields, params = table._get_select_all_sql(order_by, limit, where, fields)
        pretty = PrettyTable()
        pretty.field_names = fields
        try:
            for row in self._execute(sql, params).fetchall():
                new_fields, row = self._dereference(table, fields, row)
                data = dict(zip(new_fields, row))
                result.append(table(**data))
                pretty.add_row(list(row))
            result = tuple(result)
            if len(result) == 1: result = result[0]
            return print(pretty) if self.config['testing'] or pretty_table else result
        except AttributeError:
            pass

    def first(self, table: Union[Table]) -> Table:
        """
        Returns first row from `table`
        :params
            table: Table Object that will used
        :return:
            Table Object
        """
        sql, fields = table._get_first_sql(), table._get_column_names()
        row = self._execute(sql).fetchone()
        fields, row = self._dereference(table, fields, row)
        data = dict(zip(fields, row))
        return table(**data)

    def count(self, table: Union[Table] = None) -> int:
        """
        Returns first row from `table`
        :params
            table: Table Object that will used
            type: Type of quering data
        :return:
            int: Number of row in column or tables
        """
        if table:
            sql = table._get_count_sql()
            row = int(self._execute(sql).fetchone()[0])
            return row
        else:
            return len(self.tables)

    def get(self, table: Table, fields: list = None, **kwargs) -> Union[Table, tuple, None]:
        """
        Returns row from `table` where ROWID equals `id`
        :params
            table: Table Object that will used
            fields: List - searching fields with value
        :return:
            Table Object
        """
        try:
            sql, fields, params = table._get_select_where_sql(fields=fields, **kwargs)
            result = []
            for row in self._execute(sql, params).fetchall():
                new_fields, row = self._dereference(table, fields, row)
                data = dict(zip(new_fields, row))
                result.append(table(**data))
            if len(result) == 1:
                return result[0]
            elif len(result) == 0:
                return None
            return tuple(result)
        except Exception:
            pass

    def update(self, instance: Table):
        sql, values = instance._get_update_sql()
        self.conn.execute(sql, values)
        self.conn.commit()

    def delete(self, instance: Table):
        if isinstance(instance, Table):
            sql, params = instance._get_delete_sql(id=instance.id)
            self._execute(sql, params)
            self.conn.commit()
        else:
            raise TableTypeInvalid(instance)

    def _dereference(self, table: Table, fields: List, row: Union[List, Tuple]) -> Tuple:
        new_fields = []
        new_values = []
        for field, value in zip(fields, row):
            if field.endswith('_id'):
                # strip off "_id" to find field name
                field = field[:-3]
                fk = getattr(table, field)
                # # fetch object with the given ID
                value = self.get(fk.table, id=value)
            new_fields.append(field)
            new_values.append(value)
        return new_fields, new_values


__all__ = [
    'Database'
]



from datetime import datetime

SELECT_TABLES_SQL = "SELECT name FROM sqlite_master WHERE type = 'table';"
CREATE_TABLE_SQL = "CREATE TABLE {name} ({fields});"
INSERT_SQL = 'INSERT INTO {name} ({fields}) VALUES ({placeholders});'
SELECT_ALL_SQL = 'SELECT {fields} FROM {name}'
ALTER_MTM_SQL = 'ALTER TABLE {table} ADD COLUMN {name} INTEGER;'
SELECT_FIRST_SQL = 'SELECT * FROM {name} ORDER BY ROWID DESC LIMIT 1;'
SELECT_WHERE_SQL = 'SELECT {fields} FROM {name} WHERE {query};'
DELETE_SQL = 'DELETE FROM {name} WHERE {query};'
DROP_SQL = 'DROP TABLE {exp} {name};'
UPDATE_SQL = 'UPDATE {name} SET {fields} WHERE id = ?'

SQLITE_TYPE_MAP = {
    int: "INTEGER",
    float: "REAL",
    str: "TEXT",
    bytes: "BLOB",
    bool: "INTEGER",  # 0 or 1
    datetime: "TIMESTAMP"
}

IF_EXISTS = "IF EXISTS"

__all__ = [
    "SELECT_TABLES_SQL",
    "CREATE_TABLE_SQL",
    "INSERT_SQL",
    "SELECT_ALL_SQL",
    'ALTER_MTM_SQL',
    "SELECT_FIRST_SQL",
    "SELECT_WHERE_SQL",
    "SQLITE_TYPE_MAP",
    "DELETE_SQL",
    "DROP_SQL",
    "IF_EXISTS",
    "UPDATE_SQL",
    # "SELECT_ORDER_BY_SQL"
]



import inspect
from typing import List

from ormx.constants import *
from ormx.exceptions import *
from ormx.types import *


class Table:
    """
    Sub-class of Database, itself one table of Database
    Have same methods but without table name argument
    Attributes
    ----------
    name : str
        Name of table
    columns = list
        Generator of columns in table
    columns_names = list
        Generator of column's names in table
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        Keyword Anguments
        """
        self._data = {
            'id': None
        }
        for key, value in kwargs.items():
            self._data[key] = value

    def __getattribute__(self, key):
        _data = object.__getattribute__(self, '_data')

        if key in _data:
            return _data[key]
        return object.__getattribute__(self, key)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key in self._data:
            self._data[key] = value

    @classmethod
    def _get_name(cls):
        return cls.__tablename__.split()[0] if hasattr(cls, '__tablename__') else cls.__name__.lower()

    @classmethod
    def _get_create_sql(cls):
        fields = [
            ("id", "INTEGER PRIMARY KEY AUTOINCREMENT")
        ]
        mtm_sql = ''

        for name, field in inspect.getmembers(cls):
            if isinstance(field, Column):
                fields.append((name, field.sql_type))
            elif isinstance(field, ForeignKey):
                fields.append((name + "_id", "INTEGER"))
            elif isinstance(field, MTM):
                if not field.table._get_name() + '_id' in cls._get_column_names():
                    fields.append((name + "_id", "INTEGER"))
                mtm_sql = ALTER_MTM_SQL.format(table=field.table._get_name(),
                                               name=cls._get_name() + '_id')

        fields = [" ".join(x) for x in fields]
        sql = CREATE_TABLE_SQL.format(
            name=cls._get_name(),
            fields=", ".join(fields))

        return sql, mtm_sql

    @classmethod
    def _get_column_names(cls):
        fields = ['id']
        for name, field in inspect.getmembers(cls):
            if isinstance(field, Column):
                fields.append(name)
            if isinstance(field, ForeignKey):
                fields.append(name + "_id")
        return fields

    def _get_insert_sql(self):
        cls = self.__class__
        fields = []
        placeholders = []
        values = []

        for name, field in inspect.getmembers(cls):
            if isinstance(field, Column):
                fields.append(name)

                if field.default and isinstance(getattr(self, name), Column):
                    if type(field.default) == field.type:
                        values.append(field.default if type(getattr(self, name)).__name__ == 'Column' and type(
                            field.default) == field.type else getattr(self, name))
                    else:
                        raise TypeError(
                            f'Excepted {field.type.__name__}, but given {type(field.default).__name__} type for default value')
                else:
                    if type(getattr(self, name)) == field.type:
                        values.append(getattr(self, name))
                    else:
                        raise TypeError(
                            f'Excepted {field.type.__name__}, but given {type(getattr(self, name)).__name__} type for value')

                placeholders.append('?')
            elif isinstance(field, ForeignKey):
                fields.append(name + "_id")
                values.append(getattr(self, name).id if isinstance(getattr(self, name), Table) else 0)
                placeholders.append('?')

        sql = INSERT_SQL.format(name=cls._get_name(),
                                fields=", ".join(fields),
                                placeholders=", ".join(placeholders))

        return sql, values

    @classmethod
    def _get_first_sql(cls):
        sql = SELECT_FIRST_SQL.format(name=cls._get_name())

        return sql

    @classmethod
    def _get_count_sql(cls):
        sql = SELECT_ALL_SQL.format(fields='COUNT(*)', name=cls._get_name())

        return sql

    def _get_update_sql(self):
        cls = self.__class__
        fields = []
        values = []

        for name, field in inspect.getmembers(cls):
            if isinstance(field, Column):
                fields.append(name)
                values.append(getattr(self, name))
            elif isinstance(field, ForeignKey):
                fields.append(name + "_id")
                values.append(getattr(self, name).id)

        values.append(getattr(self, 'id'))

        sql = UPDATE_SQL.format(
            name=cls._get_name(),
            fields=', '.join([f"{field} = ?" for field in fields])
        )

        return sql, values

    @classmethod
    def _get_select_all_sql(cls,
                            order_by: tuple,
                            limit: list = None,
                            where: list = None,
                            fields: list = None
                            ):
        params = []
        fields = cls._get_column_names() if fields is None else fields
        sql = SELECT_ALL_SQL.format(name=cls._get_name(),
                                    fields=", ".join(fields))

        if where:
            if isinstance(where[0], str):
                if where[1] in WHERE_OPTS:
                    sql += f' WHERE ({where[0]} {where[1]} ?)'
                    params.append(where[2])
            elif all(isinstance(x, tuple) for x in where):
                filters = []
                sql += ' WHERE '
                for i in where:
                    if i[1] in WHERE_OPTS:
                        filters.append(f"{i[0]} {i[1]} ?")
                        params.append(i[2])
                    else:
                        raise TypeError(
                            f"Second parameter in list is wrong, it must be one of: {', '.join(WHERE_OPTS)}")
                sql += f" AND ".join(filters)

            else:
                for i in where[1::2]:
                    if not i in WHERE_CONDITIONS:
                        raise TypeError("Item's type must str or tuple")
                sql += ' WHERE '
                for i in where:
                    if isinstance(i, tuple):
                        sql += f"({i[0]} {i[1]} ?)"
                        params.append(i[2])
                    elif i in WHERE_CONDITIONS:
                        sql += f' {i} '

        if order_by:
            if not isinstance(order_by, tuple):
                raise OrderByParamError(order_by)
            if not (isinstance(order_by[0], str) and order_by[0] in fields):
                raise OrderByColumnError(order_by[0])
            if not (order_by[1] in ORDER_BY_PARAMS):
                raise SortingTypeError(order_by[1])
            sql += f' ORDER BY {order_by[0]} ?'
            params.append(order_by[1])

        if limit:
            if isinstance(limit, list):
                for i in limit:
                    if not isinstance(i, int): raise TypeError(
                        f"Parameters must be int, not {type(i).__name__}")
                if len(limit) == 1:
                    sql += f' LIMIT ?'
                    params.append(limit[0])
                elif len(limit) == 2:
                    sql += f' LIMIT ? OFFSET ?'
                    params.append(limit[0])
                    params.append(limit[1])
                else:
                    raise LimitTooMuchParamsError(limit)

        return sql, fields, tuple(params)

    @classmethod
    def _rows(cls):
        return inspect.getmembers(cls)

    @classmethod
    def _get_delete_sql(cls, **kwargs):
        filters = []
        params = []
        for key, value in kwargs.items():
            filters.append(key + " = ?")
            params.append(value)
        sql = DELETE_SQL.format(name=cls._get_name(),
                                query=" AND ".join(filters))

        return sql, tuple(params)

    @classmethod
    def _get_drop_sql(cls, exp=None):
        sql = DROP_SQL.format(name=cls._get_name(), exp=IF_EXISTS if exp else '')

        return sql

    @classmethod
    def _get_table_scheme(cls):
        sql = f"PRAGMA table_info('{cls._get_name()}')"

        return sql

    @classmethod
    def _get_select_where_sql(cls, fields: list = None, **kwargs):
        fields = fields or cls._get_column_names()

        filters = []
        params = []
        for key, value in kwargs.items():
            filters.append(key + " = ?")
            params.append(value)

        sql = SELECT_WHERE_SQL.format(name=cls._get_name(),
                                      fields=", ".join(fields),
                                      query=" AND ".join(filters))
        return sql, fields, tuple(params)

    @classmethod
    def _name(cls):
        attributes = inspect.getmembers(cls, lambda a: not (inspect.isroutine(a)))
        atrs = [a for a in attributes if not (a[0].startswith('__') and a[0].endswith('__'))]
        return atrs


class ListQuery:
    def __init__(self, db, key):
        self.db = db
        self.key = key

    def __new__(cls, key, db, *args, **kwargs) -> list:
        if key in db.tables:
            rows = db.cur.execute(f"SELECT * FROM {key}").fetchall()
            fields = [description[0] for description in db.cur.description]
            return [dict(zip(fields, rows[i])) for i in range(len(rows))]
        else:
            raise TableInfoError


class Column:

    def __init__(self, type, default=None):
        self.type = type
        self.default = default

    @property
    def sql_type(self):
        return SQLITE_TYPE_MAP[self.type]


class ForeignKey:

    def __init__(self, table):
        self.table = table


class MTM:

    def __init__(self, db, table):
        self.table = table


__all__ = [
    'Table',
    'ListQuery',
    'Column',
    'ForeignKey'
]
